/*****************************************************************************/
/*                                                                           */
/*  FILE        :papci.c                                                     */
/*  DESCRIPTION :PA10-(6C/7C) motion control board Linux Driver source file  */
/*  history:                                                                 */
/*     v1.0.0:  2002.11.30 by m.togawa                               	     */
/*        original issue                                                     */
/*                                                                           */
/*     Modifs par Matthieu Herrb pour noyaux 2.6.xx                          */
/*                                   Mars 2005                               */
/*                                                                           */
/*****************************************************************************/

#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/pci.h>

#include <asm/uaccess.h>

#undef DEBUG

MODULE_AUTHOR("M. Togawa");
MODULE_DESCRIPTION("Mitsubishi PA10 motion control board driver");
MODULE_LICENSE("Proprietary");

#include	"papci.h"

#define PAPCI_DEV_NAME "papci"

#define PAPCI_VENDOR_ID 0x8086	
#define PAPCI_DEVICE_ID 0xb555

/* XXX magic numbers */
#define PAPCI_INTR_OFFSET 0x9e
#define PAPCI_MEM_OFFSET 0x006001a0

/**
 ** Device structure 
 **/
struct papci_dev {
	int usecount;
	unsigned long PortAdr;
	unsigned char *mem;
	unsigned long mem_pci;
	unsigned long mem_len;	/* length of the region */
	unsigned char *intr; 	
	unsigned long intr_pci;
	unsigned long intr_len;	/* length of the region */
	struct cdev cdev;		/* char device */
};

/**
 ** Device numbers - defaults to dynamic assignement 
 **/
#ifndef PAPCI_MAJOR
# define PAPCI_MAJOR 0
#endif

#ifndef PAPCI_MINOR
# define PAPCI_MINOR 0
#endif

int devmajor = PAPCI_MAJOR;
int devminor = PAPCI_MINOR;

module_param(devmajor, int, S_IRUGO);
module_param(devminor, int, S_IRUGO);

/**
 ** This is the variable containing the device structure
 ** XXX this only allows for one board for now
 **/
struct papci_dev papci;

/*----------------------------------------------------------------------*/

/**
 ** probe routine
 **/
static int
papci_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	unsigned long baseaddr, endaddr;
	struct resource *res;

	printk(KERN_INFO "%s:found device(%04X,%04X) on %s\n", 
	       PAPCI_DEV_NAME,
	    PAPCI_VENDOR_ID, PAPCI_DEVICE_ID, pci_name(dev));
	if (pci_enable_device(dev) < 0) {
		printk(KERN_WARNING "%s: error enabling device\n", 
		       PAPCI_DEV_NAME);
		return -1;
	}
	/* read pci config area (interrupt mem area) */
	baseaddr = pci_resource_start(dev, 0);
	endaddr = pci_resource_end(dev, 0);
	papci.intr_len =  endaddr - baseaddr + 1;
	res = request_mem_region(baseaddr, papci.intr_len, 
				 "papci intr mem");
	if (res == NULL) {
		printk(KERN_WARNING "%s: error requesting mem region "
		       "0x%lx/0x%lx\n", PAPCI_DEV_NAME, 
		       baseaddr, papci.intr_len);
	}
	papci.intr_pci = baseaddr;
	papci.intr = ioremap(baseaddr + PAPCI_INTR_OFFSET, papci.intr_len);
	printk(KERN_INFO "\tintr addr=0x%lx / %p\n",
	       baseaddr, papci.intr);

	/* read pci config area (mem area) */
	baseaddr = pci_resource_start(dev, 4);
	endaddr = pci_resource_end(dev, 4);
	papci.mem_len =  endaddr - baseaddr + 1;
	res = request_mem_region(baseaddr, papci.mem_len, "papci mem");
	if (res == NULL) {
		printk(KERN_WARNING "%s: error requesting mem region "
		       "0x%lx/0x%lx\n", PAPCI_DEV_NAME, 
		       baseaddr, papci.mem_len);
	}
	papci.mem_pci = baseaddr;
	papci.mem = ioremap(baseaddr + PAPCI_MEM_OFFSET, papci.mem_len);
	printk(KERN_INFO "\tmem  addr=0x%lx / %p\n",
	       baseaddr, papci.mem);
	
	/* read pci config area (I/O port area) */
	papci.PortAdr = pci_resource_start(dev, 1);

	if(papci.PortAdr != 0) {
		request_region(papci.PortAdr, 0x20, "papci i/o");
	}
	printk(KERN_INFO "\tio addr=0x%lx\n", papci.PortAdr);
	return 0;
}

static void
papci_remove(struct pci_dev *dev)
{

	if (papci.mem != NULL) {
		iounmap(papci.mem);
		papci.mem = NULL;
	}
	if (papci.mem_pci != 0) {
		release_mem_region(papci.mem_pci, papci.mem_len);
		papci.mem_pci = 0;
	}
	if (papci.intr != NULL) {
		iounmap(papci.intr);
		papci.intr = NULL;
	}
	if (papci.intr_pci != 0) {
		release_mem_region(papci.intr_pci, papci.intr_len);
		papci.intr_pci = 0;
	}
	if (papci.PortAdr != 0) {
		release_region(papci.PortAdr,0x20);
		papci.PortAdr = 0;
	}
	return;
}
/*----------------------------------------------------------------------*/

/**
 ** File operations
 **/
static int 
papci_open(struct inode * inode, struct file * file)
{
	if (papci.usecount != 0) 
		return -EBUSY;
	papci.usecount++;
	return 0;
}

static int 
papci_release(struct inode * inode, struct file * file)
{

	papci.usecount--;
	return 0;
}

static int 
papci_ioctl(struct inode *inode, struct file *file, 
    unsigned int function, unsigned long arg)
{
	int stat = 0;
	PA_ARG parg;

	if (!access_ok(VERIFY_WRITE, (void*)arg, sizeof(PA_ARG))) {
		printk(KERN_NOTICE "%s access_ok error\n", __func__);
		return(-1);
	}

	if (copy_from_user(&parg, (void*)arg, sizeof(PA_ARG))) {
		printk(KERN_NOTICE "%s copy_from_user error\n", __func__);
		return -EFAULT;
	}
#ifdef DEBUG
	  printk(KERN_DEBUG "%s: arm=%ld addr=0x%lx size=%ld\n",
		 __func__, parg.arm, parg.addr, parg.size);
#endif
	switch(function){
	case PA_WRITE_PCI:
#ifdef DEBUG
		printk(KERN_DEBUG "PA_WRITE_PCI\n");
#endif
		memcpy_toio((void*)(papci.mem + parg.addr), 
		    &parg.data[0], parg.size);
		break;
	case PA_READ_PCI:
#ifdef DEBUG
		printk(KERN_DEBUG "PA_READ_PCI\n");
#endif
		memcpy_fromio((void*)&parg.data[0],
		    (void*)(papci.mem + parg.addr), parg.size);
		if (copy_to_user((void*)arg, (void*)&parg, sizeof(PA_ARG))) 
			return -EFAULT;
		break;
	case PA_INTR_PCI:
#ifdef DEBUG
		printk(KERN_DEBUG "PA_INTR_PCI\n");
#endif
		iowrite8(parg.data[0], papci.intr);
		/* memcpy(papci.intr, &parg.data[0], 1);*/
		break;
	}
	return stat;
}

/*----------------------------------------------------------------------*/

/**
 ** Module glue 
 **/

/* PCI driver attachement */

static const struct pci_device_id papci_ids[] = {
	{ PCI_DEVICE(PAPCI_VENDOR_ID, PAPCI_DEVICE_ID), },
	{ 0, },
};

MODULE_DEVICE_TABLE(pci, papci_ids);

static struct pci_driver pci_driver = { 
	.name = PAPCI_DEV_NAME,
	.id_table = papci_ids,
	.probe = papci_probe,
	.remove = papci_remove,
};

struct file_operations papci_fops = {
	.open =		papci_open,
	.release =	papci_release,
	.ioctl =	papci_ioctl,
};

static void __exit 
papci_cleanup(void)
{
	dev_t dev = MKDEV(devmajor, devminor);
	
	pci_unregister_driver(&pci_driver);
	cdev_del(&papci.cdev);
	unregister_chrdev_region(dev, 1); /* xxx */
	printk(KERN_INFO "%s: bye\n", PAPCI_DEV_NAME);
}

static int __init 
papci_init(void)
{
	int err;
	dev_t dev;

	printk(KERN_INFO "%s: initializing\n", PAPCI_DEV_NAME);
	memset(&papci, 0, sizeof(struct papci_dev));
	if ((err = pci_register_driver(&pci_driver)) < 0)
		return err;
	if (devmajor) {
		dev = MKDEV(devmajor, devminor);
		err = register_chrdev_region(dev, 1, PAPCI_DEV_NAME);
	} else {
		err = alloc_chrdev_region(&dev, 0, 1, PAPCI_DEV_NAME);
		devmajor = MAJOR(dev);
	}
	if (err < 0) {
		printk(KERN_WARNING "%s: can't get major %d\n", 
		    PAPCI_DEV_NAME, devmajor);
		return err;
	}
	printk(KERN_INFO "%s: using major number %d\n", 
	    PAPCI_DEV_NAME, devmajor);
	cdev_init(&papci.cdev, &papci_fops);
	papci.cdev.owner = THIS_MODULE;
	papci.cdev.ops = &papci_fops;
	if ((err = cdev_add(&papci.cdev, dev, 1)) < 0) {
		printk(KERN_NOTICE "Error %d adding papci\n", err);
		return err;
	}
	return 0;
}

module_init(papci_init);
module_exit(papci_cleanup);
